from __future__ import annotations

from collections.abc import Iterable
from typing import Generic, NoReturn, TypeVar

from arclet.letoderea import ParsingStop, StepOut
from satori.client.account import Account
from satori.element import Element
from satori.model import Channel, Guild, Member, MessageObject, PageResult, Role, User

from .event import Event, FriendRequestEvent, GuildMemberRequestEvent, GuildRequestEvent, MessageEvent
from .message import MessageChain

TEvent = TypeVar("TEvent", bound=Event)


class Session(Generic[TEvent]):
    """在 Satori-Python 的 Session 的基础上存储了当次事件的信息"""

    def __init__(self, account: Account, event: TEvent):
        self.account = account
        self.context = event
        self.origin = event._origin
        self._content = None
        if isinstance(event, MessageEvent):
            self._content = MessageChain(event.message.message)

    async def prompt(
        self,
        message: str | Iterable[str | Element],
        timeout: float = 120,
        timeout_message: str | Iterable[str | Element] = "等待超时",
        keep_sender: bool = True,
    ) -> MessageChain:
        """发送提示消息, 并等待回复

        Args:
            message: 要发送的消息
            timeout: 等待超时时间
            timeout_message: 超时后发送的消息
            keep_sender: 是否只允许原发送者回复
        """
        await self.send(message)

        async def waiter(content: MessageChain, session: Session[MessageEvent]):
            if self.origin.channel:
                if self.origin.channel.id == session.context.channel.id and (
                    not keep_sender
                    or (
                        self.origin.user
                        and session.context.user
                        and self.origin.user.id == session.context.user.id
                    )
                ):
                    return content
            elif self.origin.user:
                if self.origin.user.id == session.context.user.id:
                    return content

        waiter.__annotations__ = {"content": MessageChain, "session": self.__class__}

        step = StepOut([MessageEvent], waiter)

        result = await step.wait(timeout=timeout)
        if not result:
            await self.send(timeout_message)
            raise ParsingStop()
        return result

    def stop(self) -> NoReturn:
        raise ParsingStop()

    @property
    def user(self) -> User:
        if not self.origin.user:
            raise RuntimeError(f"Event {self.origin.type!r} has no User")
        return self.origin.user

    @property
    def guild(self) -> Guild:
        if not self.origin.guild:
            raise RuntimeError(f"Event {self.origin.type!r} has no Guild")
        return self.origin.guild

    @property
    def channel(self) -> Channel:
        if not self.origin.channel:
            raise RuntimeError(f"Event {self.origin.type!r} has no Channel")
        return self.origin.channel

    @property
    def member(self) -> Member:
        if not self.origin.member:
            raise RuntimeError(f"Event {self.origin.type!r} has no Member")
        return self.origin.member

    @property
    def content(self) -> str:
        if self._content:
            return str(self._content)
        raise RuntimeError(f"Event {self.origin.type!r} has no Content")

    @content.setter
    def content(self, value: str):
        self._content = MessageChain(value)

    @property
    def elements(self) -> MessageChain:
        if self._content:
            return self._content
        raise RuntimeError(f"Event {self.origin.type!r} has no Content")

    def __getattr__(self, item):
        return getattr(self.account.protocol, item)

    async def send(
        self,
        message: str | Iterable[str | Element],
    ) -> list[MessageObject]:
        return await self.account.protocol.send(self.origin, message)

    async def send_message(
        self,
        message: str | Iterable[str | Element],
    ) -> list[MessageObject]:
        """发送消息

        Args:
            message: 要发送的消息
        """
        if not self.origin.channel:
            raise RuntimeError("Event cannot be replied to!")
        return await self.account.protocol.send_message(self.origin.channel, message)

    async def send_private_message(
        self,
        message: str | Iterable[str | Element],
    ) -> list[MessageObject]:
        """发送私聊消息

        Args:
            message: 要发送的消息
        """
        if not self.origin.user:
            raise RuntimeError("Event cannot be replied to!")
        return await self.account.protocol.send_private_message(self.origin.user, message)

    async def update_message(
        self,
        message: str | Iterable[str | Element],
    ):
        """更新消息

        Args:
            message: 要更新的消息
        """
        if not self.origin.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.origin.message:
            raise RuntimeError("Event cannot update message")
        return await self.account.protocol.update_message(
            self.origin.channel, self.origin.message.id, message
        )

    async def message_create(
        self,
        content: str,
    ) -> list[MessageObject]:
        if not self.origin.channel:
            raise RuntimeError("Event cannot be replied to!")
        return await self.account.protocol.message_create(self.origin.channel.id, content)

    async def message_delete(self) -> None:
        if not self.origin.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.origin.message:
            raise RuntimeError("Event cannot update message")
        await self.account.protocol.message_delete(
            self.origin.channel.id,
            self.origin.message.id,
        )

    async def message_update(
        self,
        content: str,
    ) -> None:
        if not self.origin.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.origin.message:
            raise RuntimeError("Event cannot update message")
        await self.account.protocol.message_update(
            self.origin.channel.id,
            self.origin.message.id,
            content,
        )

    async def channel_create(self, data: Channel) -> Channel:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to create channel!")
        return await self.account.protocol.channel_create(self.origin.guild.id, data)

    async def channel_list(self, next_token: str | None = None) -> PageResult[Channel]:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to list channel!")
        return await self.account.protocol.channel_list(self.origin.guild.id, next_token)

    async def channel_update(
        self,
        data: Channel,
    ) -> None:
        if not self.origin.channel:
            raise RuntimeError("Event cannot use to update channel!")
        return await self.account.protocol.channel_update(self.origin.channel.id, data)

    async def channel_delete(self) -> None:
        if not self.origin.channel:
            raise RuntimeError("Event cannot use to delete channel!")
        return await self.account.protocol.channel_delete(
            self.origin.channel.id,
        )

    async def user_channel_create(self) -> Channel:
        if not self.origin.user:
            raise RuntimeError("Event cannot use to create user channel!")
        return await self.account.protocol.user_channel_create(
            self.origin.user.id, self.origin.guild.id if self.origin.guild else None
        )

    async def guild_member_list(self, next_token: str | None = None) -> PageResult[Member]:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to list member!")
        return await self.account.protocol.guild_member_list(self.origin.guild.id, next_token)

    async def guild_member_get(self, user_id: str | None = None) -> Member:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to get member!")
        if user_id:
            return await self.account.protocol.guild_member_get(self.origin.guild.id, user_id)
        if not self.origin.user:
            raise RuntimeError("Event cannot use to get member!")
        return await self.account.protocol.guild_member_get(self.origin.guild.id, self.origin.user.id)

    async def guild_member_kick(self, user_id: str | None = None, permanent: bool = False) -> None:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to kick member!")
        if user_id:
            return await self.account.protocol.guild_member_kick(self.origin.guild.id, user_id, permanent)
        if not self.origin.user:
            raise RuntimeError("Event cannot use to kick member!")
        return await self.account.protocol.guild_member_kick(
            self.origin.guild.id, self.origin.user.id, permanent
        )

    async def guild_member_role_set(self, role_id: str, user_id: str | None = None) -> None:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to guild member role set!")
        if user_id:
            return await self.account.protocol.guild_member_role_set(self.origin.guild.id, user_id, role_id)
        if not self.origin.user:
            raise RuntimeError("Event cannot use to guild member role set!")
        return await self.account.protocol.guild_member_role_set(
            self.origin.guild.id, self.origin.user.id, role_id
        )

    async def guild_member_role_unset(self, role_id: str, user_id: str | None = None) -> None:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to guild member role unset!")
        if user_id:
            return await self.account.protocol.guild_member_role_unset(self.origin.guild.id, user_id, role_id)
        if not self.origin.user:
            raise RuntimeError("Event cannot use to guild member role unset!")
        return await self.account.protocol.guild_member_role_unset(
            self.origin.guild.id, self.origin.user.id, role_id
        )

    async def guild_role_list(self, next_token: str | None = None) -> PageResult[Role]:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to list role!")
        return await self.account.protocol.guild_role_list(self.origin.guild.id, next_token)

    async def guild_role_create(
        self,
        role: Role,
    ) -> Role:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to create role!")
        return await self.account.protocol.guild_role_create(self.origin.guild.id, role)

    async def guild_role_update(
        self,
        role_id: str,
        role: Role,
    ) -> None:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to update role!")
        return await self.account.protocol.guild_role_update(self.origin.guild.id, role_id, role)

    async def guild_role_delete(self, role_id: str) -> None:
        if not self.origin.guild:
            raise RuntimeError("Event cannot use to delete role!")
        return await self.account.protocol.guild_role_delete(
            self.origin.guild.id,
            role_id,
        )

    async def reaction_create(
        self,
        emoji: str,
    ) -> None:
        if not self.origin.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.origin.message:
            raise RuntimeError("Event cannot create reaction")
        return await self.account.protocol.reaction_create(
            self.origin.channel.id, self.origin.message.id, emoji
        )

    async def reaction_delete(
        self,
        emoji: str,
        user_id: str | None = None,
    ) -> None:
        if not self.origin.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.origin.message:
            raise RuntimeError("Event cannot delete reaction")
        return await self.account.protocol.reaction_delete(
            self.origin.channel.id, self.origin.message.id, emoji, user_id
        )

    async def reaction_clear(
        self,
        emoji: str | None = None,
    ) -> None:
        if not self.origin.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.origin.message:
            raise RuntimeError("Event cannot clear reaction")
        return await self.account.protocol.reaction_clear(
            self.origin.channel.id,
            self.origin.message.id,
            emoji,
        )

    async def reaction_list(
        self,
        emoji: str,
        next_token: str | None = None,
    ) -> PageResult[User]:
        if not self.origin.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.origin.message:
            raise RuntimeError("Event cannot list reaction")
        return await self.account.protocol.reaction_list(
            self.origin.channel.id, self.origin.message.id, emoji, next_token
        )

    async def request_approve(
        self,
        approve: bool,
        comment: str,
    ):
        if isinstance(self.context, FriendRequestEvent):
            return await self.friend_approve(approve, comment)
        if isinstance(self.context, GuildRequestEvent):
            return await self.guild_approve(approve, comment)
        if isinstance(self.context, GuildMemberRequestEvent):
            return await self.guild_member_approve(approve, comment)
        raise RuntimeError("Event cannot approve request")

    async def friend_approve(
        self,
        approve: bool,
        comment: str,
    ):
        if not isinstance(self.context, FriendRequestEvent):
            raise RuntimeError("Event cannot approve friend request")
        return await self.account.protocol.friend_approve(self.context.message.id, approve, comment)

    async def guild_approve(
        self,
        approve: bool,
        comment: str,
    ):
        if not isinstance(self.context, GuildRequestEvent):
            raise RuntimeError("Event cannot approve guild request")
        return await self.account.protocol.guild_approve(self.context.message.id, approve, comment)

    async def guild_member_approve(
        self,
        approve: bool,
        comment: str,
    ):
        if not isinstance(self.context, GuildMemberRequestEvent):
            raise RuntimeError("Event cannot approve guild member request")
        return await self.account.protocol.guild_member_approve(self.context.message.id, approve, comment)



from __future__ import annotations

import asyncio
from contextlib import suppress

from arclet.letoderea import (
    BaseAuxiliary,
    Contexts,
    EventSystem,
    Param,
    Provider,
    ProviderFactory,
    global_providers,
)
from loguru import logger
from satori.client import App
from satori.client.account import Account
from satori.client.protocol import ApiProtocol
from satori.config import Config
from satori.model import Event
from tarina.generic import get_origin

from .command import _commands
from .event import MessageEvent, event_parse
from .plugin.model import _plugins
from .session import Session


class ApiProtocolProvider(Provider[ApiProtocol]):
    async def __call__(self, context: Contexts):
        if account := context.get("$account"):
            return account.protocol


class SessionProvider(Provider[Session]):
    def validate(self, param: Param):
        return get_origin(param.annotation) == Session

    async def __call__(self, context: Contexts):
        if "$origin_event" in context and "$account" in context:
            return Session(context["$account"], context["$event"])


global_providers.extend([ApiProtocolProvider(), SessionProvider()])


class Entari(App):
    id = "entari.service"

    def __init__(self, *configs: Config):
        super().__init__(*configs)
        self.event_system = EventSystem()
        self.event_system.register(_commands.publisher)
        self.register(self.handle_event)
        self._ref_tasks = set()

    def on(
        self,
        *events: type,
        priority: int = 16,
        auxiliaries: list[BaseAuxiliary] | None = None,
        providers: list[Provider | type[Provider] | ProviderFactory | type[ProviderFactory]] | None = None,
    ):
        return self.event_system.on(*events, priority=priority, auxiliaries=auxiliaries, providers=providers)

    def on_message(
        self,
        priority: int = 16,
        auxiliaries: list[BaseAuxiliary] | None = None,
        providers: list[Provider | type[Provider] | ProviderFactory | type[ProviderFactory]] | None = None,
    ):
        return self.event_system.on(
            MessageEvent, priority=priority, auxiliaries=auxiliaries, providers=providers
        )

    async def handle_event(self, account: Account, event: Event):
        async def event_parse_task(connection: Account, raw: Event):
            loop = asyncio.get_running_loop()
            with suppress(NotImplementedError):
                ev = event_parse(connection, raw)
                self.event_system.publish(ev)
                for plugin in _plugins.values():
                    for disp in plugin.dispatchers.values():
                        if not disp.validate(ev):
                            continue
                        if disp._run_by_system:
                            continue
                        task = loop.create_task(disp.publish(ev))
                        self._ref_tasks.add(task)
                        task.add_done_callback(self._ref_tasks.discard)
                return

            logger.warning(f"received unsupported event {raw.type}: {raw}")

        await event_parse_task(account, event)



from typing import Optional

from arclet.letoderea import Interface, JudgeAuxiliary, Scope
from satori import Channel, ChannelType


class DirectMessageJudger(JudgeAuxiliary):
    async def __call__(self, scope: Scope, interface: Interface) -> Optional[bool]:
        if not (channel := interface.query(Channel, "channel", force_return=True)):
            return False
        return channel.type == ChannelType.DIRECT

    @property
    def scopes(self) -> set[Scope]:
        return {Scope.prepare}


is_direct_message = DirectMessageJudger()


class PublicMessageJudger(JudgeAuxiliary):
    async def __call__(self, scope: Scope, interface: Interface) -> Optional[bool]:
        if not (channel := interface.query(Channel, "channel", force_return=True)):
            return False
        return channel.type != ChannelType.DIRECT

    @property
    def scopes(self) -> set[Scope]:
        return {Scope.prepare}


is_public_message = PublicMessageJudger()



from arclet.alconna import Alconna, command_manager

from ..event import MessageEvent
from ..plugin import Plugin, PluginDispatcher
from .model import Match, Query
from .provider import AlconnaProviderFactory, AlconnaSuppiler, MessageJudger


class AlconnaPluginDispatcher(PluginDispatcher):

    def __init__(
        self,
        plugin: Plugin,
        command: Alconna,
        need_tome: bool = False,
        remove_tome: bool = False,
    ):
        self.supplier = AlconnaSuppiler(command, need_tome, remove_tome)
        super().__init__(plugin, MessageEvent)

        self.bind(MessageJudger(), self.supplier)
        self.bind(AlconnaProviderFactory())

    def dispose(self):
        super().dispose()
        command_manager.delete(self.supplier.cmd)
        del self.supplier.cmd
        del self.supplier

    Match = Match
    Query = Query


def mount(cmd: Alconna, need_tome: bool = False, remove_tome: bool = False):
    if not (plugin := Plugin.current()):
        raise LookupError("no plugin context found")
    disp = AlconnaPluginDispatcher(plugin, cmd, need_tome, remove_tome)
    return disp



from copy import deepcopy
import inspect
from typing import Any, Literal, Optional, Union, get_args

from arclet.alconna import Alconna, Arparma, Duplication, Empty, output_manager
from arclet.alconna.builtin import generate_duplication
from arclet.letoderea import (
    Contexts,
    Interface,
    JudgeAuxiliary,
    Param,
    Provider,
    Scope,
    Subscriber,
    SupplyAuxiliary,
)
from arclet.letoderea.provider import ProviderFactory
from nepattern.util import CUnionType
from satori.client import Account
from satori.element import At, Text
from tarina.generic import get_origin

from ..message import MessageChain
from .model import CommandResult, Match, Query


def _is_tome(message: MessageChain, account: Account):
    if message and isinstance(message[0], At):
        at: At = message[0]  # type: ignore
        if at.id and at.id == account.self_id:
            return True
    return False


def _remove_tome(message: MessageChain, account: Account):
    if _is_tome(message, account):
        message = deepcopy(message)
        message.pop(0)
        if message and isinstance(message[0], Text):
            text = message[0].text.lstrip()  # type: ignore
            if not text:
                message.pop(0)
            else:
                message[0] = Text(text)
        return message
    return message


class MessageJudger(JudgeAuxiliary):
    async def __call__(self, scope: Scope, interface: Interface) -> Optional[bool]:
        return "$message_content" in interface.ctx

    @property
    def scopes(self) -> set[Scope]:
        return {Scope.prepare}


class AlconnaSuppiler(SupplyAuxiliary):
    cmd: Alconna
    need_tome: bool
    remove_tome: bool

    def __init__(self, cmd: Alconna, need_tome: bool, remove_tome: bool):
        super().__init__()
        self.cmd = cmd
        self.need_tome = need_tome
        self.remove_tome = remove_tome

    async def __call__(self, scope: Scope, interface: Interface) -> Optional[Union[bool, Interface.Update]]:
        account: Account = interface.ctx["$account"]
        message: MessageChain = interface.ctx["$message_content"]
        if self.need_tome and not _is_tome(message, account):
            return False
        with output_manager.capture(self.cmd.name) as cap:
            output_manager.set_action(lambda x: x, self.cmd.name)
            if self.remove_tome:
                message = _remove_tome(message, account)
            try:
                _res = self.cmd.parse(message)
            except Exception as e:
                _res = Arparma(self.cmd.path, message, False, error_info=e)
            may_help_text: Optional[str] = cap.get("output", None)
        if _res.matched:
            return interface.update(alc_result=CommandResult(self.cmd, _res, may_help_text))
        elif may_help_text:
            await account.send(interface.event, MessageChain(may_help_text))
            return False
        return False

    @property
    def scopes(self) -> set[Scope]:
        return {Scope.prepare}


class AlconnaProvider(Provider[Any]):
    def __init__(self, type_: str, extra: Optional[dict] = None):
        super().__init__()
        self.type = type_
        self.extra = extra or {}

    async def __call__(self, context: Contexts):
        if "alc_result" not in context:
            return
        result: CommandResult = context["alc_result"]
        if self.type == "result":
            return result
        if self.type == "arparma":
            return result.result
        if self.type == "alconna":
            return result.source
        if self.type == "default_duplication":
            return generate_duplication(result.source)(result.result)
        if self.type == "duplication":
            return self.extra["duplication"](result.result)
        if self.type == "match":
            target = result.result.all_matched_args.get(self.extra["name"], Empty)
            return Match(target, target != Empty)
        if self.type == "query":
            q = Query(self.extra["query"].path, self.extra["query"].result)
            res = result.result.query(q.path, Empty)
            q.available = res != Empty
            if q.available:
                q.result = res
            elif self.extra["query"].result != Empty:
                q.available = True
            return q
        if self.extra["name"] in result.result.all_matched_args:
            return result.result.all_matched_args[self.extra["name"]]


class AlconnaProviderFactory(ProviderFactory):
    def validate(self, param: Param):
        annotation = get_origin(param.annotation)
        if annotation in (Union, CUnionType, Literal):
            annotation = get_origin(get_args(param.annotation)[0])
        if annotation is CommandResult:
            return AlconnaProvider("result")
        if annotation is Arparma:
            return AlconnaProvider("arparma")
        if annotation is Alconna:
            return AlconnaProvider("alconna")
        if annotation is Duplication:
            return AlconnaProvider("default_duplication")
        if inspect.isclass(annotation) and issubclass(annotation, Duplication):
            return AlconnaProvider("duplication", {"duplication": param.annotation})
        if annotation is Match:
            return AlconnaProvider("match", {"name": param.name})
        if isinstance(param.default, Query):
            return AlconnaProvider("query", {"query": param.default})
        return AlconnaProvider("args", {"name": param.name, "anno": param.annotation})


def get_cmd(target: Subscriber):
    return next(a for a in target.auxiliaries[Scope.prepare] if isinstance(a, AlconnaSuppiler)).cmd



from __future__ import annotations

from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Callable

from loguru import logger

from .model import Plugin, PluginDispatcher
from .model import PluginMetadata as PluginMetadata
from .model import _current_plugin, _plugins
from .module import import_plugin

if TYPE_CHECKING:
    from ..event import Event


def dispatch(*events: type[Event], predicate: Callable[[Event], bool] | None = None):
    if not (plugin := _current_plugin.get()):
        raise LookupError("no plugin context found")
    disp = PluginDispatcher(plugin, *events, predicate=predicate)
    return disp


def load_plugin(path: str) -> Plugin | None:
    """
    以导入路径方式加载模块

    Args:
        path (str): 模块路径
    """
    if path in _plugins:
        return _plugins[path]
    try:
        mod = import_plugin(path)
        if not mod:
            logger.error(f"cannot found plugin {path!r}")
            return
        logger.success(f"loaded plugin {path!r}")
        return mod.__plugin__
    except Exception as e:
        logger.error(f"failed to load plugin {path!r} caused by {e!r}")


def load_plugins(dir_: str | PathLike | Path):
    path = dir_ if isinstance(dir_, Path) else Path(dir_)
    if path.is_dir():
        for p in path.iterdir():
            if p.suffix in (".py", "") and p.stem not in {"__init__", "__pycache__"}:
                load_plugin(".".join(p.parts[:-1:1]) + "." + p.stem)
    elif path.is_file():
        load_plugin(".".join(path.parts[:-1:1]) + "." + path.stem)


def dispose(plugin: str):
    if plugin not in _plugins:
        return
    _plugin = _plugins[plugin]
    _plugin.dispose()



from collections.abc import Sequence
from importlib.abc import MetaPathFinder
from importlib.machinery import PathFinder, SourceFileLoader
from importlib.util import module_from_spec, resolve_name
import sys
from types import ModuleType
from typing import Optional

from .model import Plugin, PluginMetadata, _current_plugin, _plugins


class PluginFinder(MetaPathFinder):
    def find_spec(
        self,
        fullname: str,
        path: Optional[Sequence[str]],
        target: Optional[ModuleType] = None,
    ):
        module_spec = PathFinder.find_spec(fullname, path, target)
        if not module_spec:
            return
        module_origin = module_spec.origin
        if not module_origin:
            return

        module_spec.loader = PluginLoader(fullname, module_origin)
        return module_spec


class PluginLoader(SourceFileLoader):
    def __init__(self, fullname: str, path: str) -> None:
        self.loaded = False
        super().__init__(fullname, path)

    def create_module(self, spec) -> Optional[ModuleType]:
        if self.name in _plugins:
            self.loaded = True
            return _plugins[self.name].module
        return super().create_module(spec)

    def exec_module(self, module: ModuleType) -> None:
        if self.loaded:
            return

        # create plugin before executing
        plugin = Plugin(module.__name__, module)
        setattr(module, "__plugin__", plugin)

        # enter plugin context
        _plugin_token = _current_plugin.set(plugin)

        try:
            super().exec_module(module)
        # except Exception:
        #     # _revert_plugin(plugin)
        #     raise
        finally:
            # leave plugin context
            _current_plugin.reset(_plugin_token)

        # get plugin metadata
        metadata: Optional[PluginMetadata] = getattr(module, "__plugin_metadata__", None)
        plugin.metadata = metadata
        return


_finder = PluginFinder()


def find_spec(name, package=None):
    fullname = resolve_name(name, package) if name.startswith(".") else name
    parent_name = fullname.rpartition(".")[0]
    if parent_name:
        parent = __import__(parent_name, fromlist=["__path__"])
        try:
            parent_path = parent.__path__
        except AttributeError as e:
            raise ModuleNotFoundError(
                f"__path__ attribute not found on {parent_name!r} " f"while trying to find {fullname!r}",
                name=fullname,
            ) from e
    else:
        parent_path = None
    return _finder.find_spec(fullname, parent_path)


def import_plugin(name, package=None):
    spec = find_spec(name, package)
    if spec:
        mod = module_from_spec(spec)
        if spec.loader:
            spec.loader.exec_module(mod)
        return mod
    return


sys.meta_path.insert(0, _finder)



from __future__ import annotations

from contextvars import ContextVar
from dataclasses import dataclass, field
from types import ModuleType
from typing import TYPE_CHECKING, Callable
from weakref import WeakValueDictionary, finalize

from arclet.letoderea import BaseAuxiliary, Provider, Publisher, StepOut, system_ctx
from arclet.letoderea.builtin.breakpoint import R
from arclet.letoderea.typing import TTarget

if TYPE_CHECKING:
    from ..event import Event

_current_plugin: ContextVar[Plugin | None] = ContextVar("_current_plugin", default=None)

_plugins: dict[str, Plugin] = {}


class PluginDispatcher(Publisher):
    def __init__(
        self,
        plugin: Plugin,
        *events: type[Event],
        predicate: Callable[[Event], bool] | None = None,
    ):
        super().__init__(f"{plugin.id}@{id(plugin)}", *events, predicate=predicate)  # type: ignore
        self.plugin = plugin
        plugin.dispatchers[self.id] = self
        self._run_by_system = False
        if es := system_ctx.get():
            es.register(self)
            self._run_by_system = True
        self._events = events

    def waiter(
        self,
        *events: type[Event],
        providers: list[Provider | type[Provider]] | None = None,
        auxiliaries: list[BaseAuxiliary] | None = None,
        priority: int = 15,
        block: bool = False,
    ) -> Callable[[TTarget[R]], StepOut[R]]:
        def wrapper(func: TTarget[R]):
            nonlocal events
            if not events:
                events = self._events
            return StepOut(list(events), func, providers, auxiliaries, priority, block)  # type: ignore

        return wrapper

    def dispose(self):
        if self._run_by_system:
            if es := system_ctx.get():
                es.publishers.pop(self.id, None)
            self._run_by_system = False
        self.subscribers.clear()

    on = Publisher.register
    handle = Publisher.register
    __call__ = Publisher.register


@dataclass
class PluginMetadata:
    name: str
    author: list[str] = field(default_factory=list)
    version: str | None = None
    license: str | None = None
    urls: dict[str, str] | None = None
    description: str | None = None
    icon: str | None = None
    classifier: list[str] = field(default_factory=list)
    dependencies: list[str] = field(default_factory=list)

    # standards: list[str] = field(default_factory=list)
    # frameworks: list[str] = field(default_factory=list)
    # config_endpoints: list[str] = field(default_factory=list)
    # component_endpoints: list[str] = field(default_factory=list)


@dataclass
class Plugin:
    id: str
    module: ModuleType
    dispatchers: WeakValueDictionary[str, PluginDispatcher] = field(default_factory=WeakValueDictionary)
    metadata: PluginMetadata | None = None
    _is_disposed: bool = False

    @staticmethod
    def current() -> Plugin | None:
        return _current_plugin.get()

    def __post_init__(self):
        _plugins[self.id] = self
        finalize(self, self.dispose)

    def dispose(self):
        if self._is_disposed:
            return
        self._is_disposed = True
        for disp in self.dispatchers.values():
            disp.dispose()
        self.dispatchers.clear()
        del _plugins[self.id]
        del self.module
